/* ISC license. */

#include <sys/types.h>
#include <sys/stat.h>
#include <string.h>
#include <limits.h>
#include <signal.h>
#include <regex.h>
#include <libgen.h>
#include <skalibs/allreadwrite.h>
#include <skalibs/bytestr.h>
#include <skalibs/buffer.h>
#include <skalibs/strerr2.h>
#include <skalibs/sgetopt.h>
#include <skalibs/sig.h>
#include <skalibs/selfpipe.h>
#include <skalibs/tai.h>
#include <skalibs/djbunix.h>
#include <skalibs/iopause.h>
#include <skalibs/skamisc.h>

#define USAGE "mdevd [ -v verbosity ] [ -f conffile ] [ -n ] [ -s slashsys ] [ -d slashdev ]"
#define dieusage() strerr_dieusage(100, USAGE)

#define BUFSIZE 8192
#define UEVENT_MAX_SIZE 4096
#define UEVENT_MAX_VARS 64

#define ACTION_ADD 0
#define ACTION_REMOVE 1

static int flagdryrun = 0 ;
static int cont = 1 ;
static pid_t pid = 0 ;
static unsigned int verbosity = 1 ;
static char const *slashsys = "/sys" ;
static char const *slashdev = "/dev" ;

struct envmatch_s
{
  unsigned short var ;
  regex_t re ;
} ;

struct majmin_s
{
  unsigned int maj ;
  unsigned int minlo ;
  unsigned int minhi ;
} ;

union devmatch_u
{
  regex_t devre ;
  struct majmin_s majmin ;
} ;

#define DEVMATCH_NOTHING 0
#define DEVMATCH_CATCHALL 1
#define DEVMATCH_DEVRE 2
#define DEVMATCH_MAJMIN 3

#define MOVEINFO_NOTHING 0
#define MOVEINFO_NOCREATE 1
#define MOVEINFO_MOVE 2
#define MOVEINFO_MOVEANDLINK 3

typedef struct scriptelem_s scriptelem, *scriptelem_ref ;
struct scriptelem_s
{
  uid_t uid ;
  gid_t gid ;
  mode_t mode ;
  union devmatch_u devmatch ;
  unsigned short envmatchs ;
  unsigned short envmatchlen : 11 ;
  unsigned short devmatchtype : 2 ;
  unsigned short movetype : 2 ;
  unsigned short flagcont : 1 ;
} ;
#define SCRIPTELEM_ZERO { .uid = 0, .gid = 0, .mode = 0, .envmatchs = 0, .moveinfo_before = 0, .moveinfo_after = 0, .envmatchlen = 0, .devmatchtype = DEVMATCH_NOTHING, .movetype = MOVEINFO_NOTHING, .flagcont = 0 }

struct uevent_s
{
  unsigned short len ;
  unsigned short varn ;
  unsigned short vars[UEVENT_MAX_VARS] ;
  char buf[UEVENT_MAX_SIZE] ;
} ;
#define UEVENT_ZERO { .len = 0, .varn = 0 }


static inline void scriptelem_free (scriptelem *elem, struct envmatch_s *envmatch)
{
  while (elem->envmatchlen--) regfree(&envmatch[elem->envmatchs++].re) ;
  if (elem->devmatchtype == DEVMATCH_DEVRE) regfree(&elem->devmatch.devre) ;
}

static inline void script_free (scriptelem *script, unsigned short scriptlen, struct envmatch_s *envmatch)
{
  while (scriptlen--) scriptelem_free(script++, envmatch) ;
}

static inline void script_firstpass (char const *s, size_t len, unsigned short *scriptlen, unsigned short *envmatchlen)
{
}

static inline void script_secondpass (char const *s, size_t len, scriptelem *script, struct envmatch_s *envmatch)
{
}

static inline int wait_for_loading (char const *sysdevpath, size_t sysdevpathlen)
{
  int lfd = -1 ;
  unsigned int n = 150 ;
  static tain_t const period = { .sec = TAI_ZERO, .nano = 200000000 } ;
  char loadingfn[sysdevpathlen + 9] ;
  memcpy(loadingfn, sysdevpath, sysdevpathlen) ;
  memcpy(loadingfn + sysdevpathlen, "/loading", 9) ;
  while (count--)  /* sysfs doesn't support inotify, so we have to poll -_- */
  {
    tain_t deadline ;
    lfd = open_writeb(loadingfn) ;
    if (lfd >= 0) break ;
    tain_add_g(&deadline, &period) ;
    deepsleepuntil_g(&deadline) ;
  }
  return lfd ;
}

static inline void load_firmware (char const *fwbase, char const *fw, char const *sysdevpath)
{
  size_t fwbaselen = strlen(fwbase) ;
  size_t fwlen = strlen(fw) ;
  size_t sysdevpathlen = strlen(sysdevpath) ;
  int fwfd, loadingfd ;
  char fwfn[fwbaselen + fwlen + 2] ;
  memcpy(fwfn, fwbase, fwbaselen) ;
  fwfn[fwbaselen] = '/' ;
  memcpy(fwfn + fwbaselen + 1, fw, fwlen + 1) ;
  fwfd = open_readb(fwfn) ;
  if (fwfd < 0 && verbosity >= 2) strerr_warnwu3sys("open ", fwfn, " for reading") ;
  loadingfd = wait_for_loading(sysdevpath, sysdevpathlen) ;
  if (loadingfd < 0)
  {
    if (verbosity >= 2) strerr_warnwu3sys("open ", sysdevpath, "/loading for writing") ;
    goto errclosef ;
  }
  if (fwfd < 0) goto errclosel ;
  if (fd_write(loadingfd, "1", 1) < 1)
  {
    if (verbosity >= 2) strerr_warnwu3sys("write 1 to ", sysdevpath, "/loading") ;
    goto errclosel ;
  }
  {
    int datafd ;
    char datafn[sysdevpathlen + 6] ;
    memcpy(datafn, sysdevpath, sysdevpathlen) ;
    memcpy(datafn + sysdevpathlen, "/data", 6) ;
    datafd = open_writeb(datafn) ;
    if (datafd < 0)
    {
      if (verbosity >= 2) strerr_warnwu3sys("open ", datafn, " for writing") ;
      goto errload ;
    }
    if (fd_cat(fwfd, datafd) < 0)
    {
      if (verbosity >= 2) strerr_warnwu4sys("copy ", fwfn, " to ", datafn) ;
      goto errload ;
    }
    fd_close(datafd) ;
    fd_write(loadingfd, "0", 1) ;
  }
  fd_close(loadingfd) ;
  fd_close(fwfd) ;
  return ;

 errload:
  allwrite(loadingfd, "-1", 2) ;
 errclosel:
  fd_close(loadingfd) ;
 errclosef:
  if (fwfd >= 0) fd_close(fwfd) ;
}

static inline int handle_signals (void)
{
  for (;;)
  {
    int c = selfpipe_read() ;
    switch (c)
    {
      case -1 : strerr_diefu1sys(111, "selfpipe_read") ;
      case 0 : return 0 ;
      case SIGHUP : return 1 ;
      case SIGCHLD :
        if (!pid) wait_reap() ;
        else
        {
          int wstat ;
          int r = wait_pid_nohang(pid, &wstat) ;
          if (r < 0)
            if (errno != ECHILD) strerr_diefu1sys(111, "wait_pid_nohang") ;
            else break ;
          else if (!r) break ;
          pid = 0 ;
        }
        break ;
      default :
        strerr_dief1x(101, "internal error: inconsistent signal handling. Please submit a bug-report.") ;
    }
  }
}

static char const *event_getvar (struct uevent_s const *event, char const *var)
{
  size_t varlen = strlen(var) ;
  unsigned short i = 1 ;
  for (; i < event->varn ; i++)
    if (!strncmp(var, event->buf + event->vars[i], varlen) && event->buf[event->vars[i] + varlen] == '=')
      break ;
  return i < event->varn ? event->buf + event->vars[i] + varlen + 1 ; 0 ;
}

static inline unsigned char class (char c)
{
  static unsigned char const classtable[64] = "0333333333333333333333333333333333333133333333332222222222333333" ;
  return (unsigned char)c < 64 ? classtable[c] : (unsigned char)'3' ;
}

static ssize_t alias_format (char *out, size_t outmax, char const *in, char const *data, regmatch_t const *off)
{
  static const table[2][4] = { { 0x12, 0x01, 0x10, 0x10 }, { 0x03, 0x10, 0x20, 0x03 } } ;
  size_t w = 0 ;
  unsigned int state = 0 ;
  while (state < 2)
  {
    char next = *in++ ;
    char what = table[state][class(next)-'0'] ;
    state = what & 0x03 ;
    if (what & 0x01)
    {
      if (w >= outmax) return -1 ;
      if (out) out[w] = next ;
      w++ ;
    }
    if (what & 0x02)
    {
      unsigned int i = next - '0' ;
      size_t len = off[i].rm_eo - off[i].rm_so ;
      if (w + len > outmax) return -1 ;
      if (out) memcpy(out + w, data + off[i].rm_so, len) ;
      w += len ;
    }
  }
  if (state == 3) return -2 ;
  return w ;
}

static inline int run_scriptelem (struct uevent_s const *event, scriptelem const *elem, char const *storage, struct envmatch_s const *envmatch, char const *devname, mode_t devtype, int mmaj, int mmin)
{
  unsigned short i = 0 ;
  regmatch_t off[10] ;
  for (; i < elem->envmatchlen ; i++)
  {
    int r ;
    char const *x = event_getvar(event, storage + envmatch[elem->envmatchs].var) ;
    if (!x) return 0 ;
    if (regexec(&envmatch[elem->envmatchs].re, x, 0, 0, 0)) return 0 ;
  }

  switch (elem->devmatchtype)
  {
    case DEVMATCH_NOTHING: break ;
    case DEVMATCH_MAJMIN:
      if (mmaj >= 0 && mmin >= 0 && mmaj == elem->devmatch.majmin.maj && mmin >= elem->devmatch.majmin.minlo && mmin <= elem->devmatch.majmin.minhi) break ;
      return 0 ;
    case DEVMATCH_DEVRE:
      if (!regexec(&elem->devmatch.devre, devname, 10, off, 0)
       && !off[0].rm_so && off[0].rm_eo == strlen(devname))
        break ;
      return 0 ;
    case DEVMATCH_CATCHALL: break ;
  }

  if (elem->movetype != MOVEINFO_NOCREATE)
  {
    char buf[PATH_MAX] ;
    size_t slashdevlen = strlen(slashdev) ;
    size_t w = slashdevlen+1 ;
    ssize_t r ;
    memcpy(buf, slashdev, w) ;
    
  }
}

static inline void run_script (struct uevent_s const *event, scriptelem const *script, unsigned short scriptlen, char const *storage, struct envmatch_s const *envmatch, char const *devname, mode_t devtype, int mmaj, int mmin)
{
  unsigned short i = 0 ;
  for (; i < scriptlen ; i++)
  {
    int r = run_scriptelem(event, script + i, storage, envmatch, devname, devtype, mmaj, mmin) ;
    if (r > 0) break ;
  }
}

static inline void doit (struct uevent_s const *event, char const *sysdevpath, size_t sysdevpathlen, unsigned int action, scriptelem const *script, unsigned short scriptlen, char const *storage, struct envmatch_s const *envmatch)
{
  ssize_t hasmajmin = 0 ;
  unsigned int mmaj, mmin ;
  mode_t devtype = S_IFCHR ;
  char const *devname ;
  char const *x = event_getvar(event, "MAJOR") ;
  char buf[UEVENT_MAX_SIZE] ;
  if (action == ACTION_ADD)
  {
    if (x && uint0_scan(x, &mmaj))
    {
      x = event_getvar(event, "MINOR") ;
      if (x && uint0_scan(x, &mmin)) hasmajmin = 1 ;
    }
    if (!hasmajmin)
    {
      memcpy(sysdevpath + sysdevpathlen, "/dev", 5) ;
      hasmajmin = openreadnclose(sysdevpath, buf, UINT_FMT << 1) ;
      sysdevpath[sysdevpathlen] = 0 ;
      if (hasmajmin > 0)
      {
        size_t i = uint_scan(buf, &mmaj) ;
        if (i > 0 && buf[i] == ':')
        {
          size_t j = uint_scan(buf + i + 1, &mmin) ;
          if (j > 0 && buf[i+1+j] == '\n') ;
          else hasmajmin = 0 ;
        }
        else hasmajmin = 0 ;
      }
    }
  }

  devname = event_getvar(event, "DEVNAME") ;
  if (!devname)
  {
    ssize_t r ;
    memcpy(sysdevpath + sysdevpathlen, "/uevent", 8) ;
    r = openreadclose(sysdevpath, buf, UEVENT_MAX_SIZE-1) ;
    sysdevpath[sysdevpathlen] = 0 ;
    if (r > 0)
    {
      buf[r] = 0 ;
      devname = strstr(buf, "\nDEVNAME=") ;
      if (devname)
      {
        devname += 9 ;
        *strchr(devname, '\n') = 0 ;
      }
    }
    if (!devname) devname = basename(sysdevpath) ;
  }

  if (strstr(sysdevpath, "/block/")) devtype = S_IFBLK ;
  else
  {
    x = event_getvar(event, "SUBSYSTEM") ;
    if (x && str_start(x, "block")) devtype = S_IFBLK ;
  }
  run_script(event, script, scriptlen, storage, envmatch, devname, devtype, hasmajmin > 0 ? mmaj : -1, hasmajmin > 0 ? mmin : -1) ;
}

static inline void on_event (struct uevent_s const *event, scriptelem const *script, unsigned short scriptlen, char const *storage, struct envmatch_s const *envmatch)
{
  unsigned int action ;
  char const *x = event_getvar(event, "ACTION") ;
  if (!x) return ;
  if (!strcmp(x, "add")) action = ACTION_ADD ;
  else if (!strcmp(x, "remove")) action = ACTION_REMOVE ;
  else return ;
  x = event_getvar(event, "DEVPATH") ;
  if (!x) return ;
  {
    size_t devpathlen = strlen(x) ;
    size_t slashsyslen = strlen(slashsys) ;
    char sysdevpath[devpathlen + slashsyslen + 2 + 7] ;
    memcpy(sysdevpath, slashsys, slashsyslen) ;
    sysdevpath[slashsyslen] = '/' ;
    memcpy(sysdevpath + slashsyslen + 1, x, devpathlen + 1) ;
    x = event_getvar(event, "FIRMWARE") ;
    if (action == ACTION_ADD || !x) doit(event, sysdevpath, slashsyslen + devpathlen + 1, action, script, scriptlen, storage, envmatch) ;
    if (action == ACTION_ADD && x) load_firmware(event, x) ;
  }
}

static void handle_stdin (struct uevent_s *event, scriptelem const *script, unsigned short scriptlen, char const *storage, struct envmatch_s const *envmatch)
{
  size_t w = 0 ;
  while (!pid)
  {
    ssize_t r = sanitize_read(getlnmax(buffer_0, event->buf + event->len, UEVENT_MAX_SIZE - event->len, &w, '\0')) ;
    if (r < 0)
    {
      cont = 0 ;
      if (errno != EPIPE && verbosity) strerr_warnwu1sys("read from stdin") ;
      if (event->len) strerr_dief1x(1, "received incomplete event") ;
    }
    if (r <= 0) break ;
    if (w > 1)
    {
      event->vars[event->varn++] = event->len ;
      event->len += w ;
    }
    else
    {
      if (event->vars > 1)
        on_event(event, script, scriptlen, storage, envmatch) ;
      event->len = 0 ;
    }
  }
}

int main (int argc, char const *const *argv)
{
  char const *configfile = "/etc/mdev.conf" ;
  iopause_fd x[2] = { { .events = IOPAUSE_READ }, { .fd = 0 } } ;
  struct uevent_s event = UEVENT_ZERO ;
  PROG = "mdevd" ;
  {
    subgetopt_t l = SUBGETOPT_ZERO ;
    for (;;)
    {
      int opt = subgetopt_r(argc, argv, "nv:f:s:d:", &l) ;
      if (opt == -1) break ;
      switch (opt)
      {
        case 'n' : flagdryrun = 1 ; break ;
        case 'v' : if (!uint0_scan(l.arg, &verbosity)) dieusage() ; break ;
        case 'f' : configfile = l.arg ; break ;
        case 's' : slashsys = l.arg ; break ;
        case 'd' : slashdev = l.arg ; break ;
        default : dieusage() ;
      }
    }
    argc -= l.ind ; argv += l.ind ;
  }

  if (strlen(slashdev) >= PATH_MAX || strlen(slashsys) >= PATH_MAX)
    strerr_diefu1x(100, "paths too long") ;
  if (ndelay_on(0) < 0) strerr_diefu1sys(111, "set stdin nonblocking") ;
  x[0].fd = selfpipe_init() ;
  if (x[0].fd < 0) strerr_diefu1sys(111, "init selfpipe") ;
  if (sig_ignore(SIGPIPE) < 0) strerr_diefu1sys(111, "ignore SIGPIPE") ;
  {
    sigset_t set ;
    sigemptyset(&set) ;
    sigaddset(&set, SIGCHLD) ;
    sigaddset(&set, SIGHUP) ;
    if (selfpipe_trapset(&set) < 0)
      strerr_diefu1sys(111, "trap signals") ;
  }
  umask(0) ;

  while (cont || pid)
  {
    unsigned short scriptlen = 0 ;
    unsigned short envmatchlen = 0 ;
    ssize_t len ;
    char buf[BUFSIZE] ;
    len = openreadnclose(configfile, buf, BUFSIZE-1) ;
    if (len < 0 && errno != ENOENT) strerr_diefu2sys(111, "read ", configfile) ;
    if (len > 0) script_firstpass(buf, len, &scriptlen, &envmatchlen) ;
    {
      scriptelem script[scriptlen ? scriptlen : 1] ;
      struct envmatch_s envmatch[envmatchlen ? envmatchlen : 1] ;
      if (scriptlen) script_secondpass(buf, len, script, envmatch) ;
      while (cont || pid)
      {
        if (buffer_len(buffer_0) handle_stdin(&event, script, scriptlen, buf, envmatch) ;
        x[1].events = pid ? 0 : IOPAUSE_READ ;
        if (iopause_g(x, 1 + cont, &tain_infinite) < 0) strerr_diefu1sys(111, "iopause") ;
        if (x[0].revents & IOPAUSE_READ && handle_signals()) break ;
        if (cont && !pid && x[1].revents & IOPAUSE_READ)
          handle_stdin(&event, script, scriptlen, buf, envmatch) ;
      }
      script_free(script, scriptlen, envmatch) ;
    }
  } 
  return 0 ;
}
